﻿using System;
using System.Runtime.CompilerServices;
using LibHac.Common;
using LibHac.Fs;
using LibHac.Fs.Fsa;
using LibHac.Os;
using LibHac.Util;

namespace LibHac.FsSystem
{
    internal struct DirectorySaveDataFileSystemGlobals
    {
        public SdkMutexType SynchronizeDirectoryMutex;

        public void Initialize(FileSystemClient fsClient)
        {
            SynchronizeDirectoryMutex.Initialize();
        }
    }

    /// <summary>
    /// An <see cref="IFileSystem"/> that provides transactional commits for savedata on top of another base IFileSystem.
    /// </summary>
    /// <remarks>
    /// Transactional commits should be atomic as long as the <see cref="IFileSystem.RenameDirectory"/> function of the
    /// underlying <see cref="IFileSystem"/> is atomic.
    /// <br/>Based on FS 11.0.0 (nnSdk 11.4.0)
    /// </remarks>
    public class DirectorySaveDataFileSystem : IFileSystem
    {
        private const int IdealWorkBufferSize = 0x100000; // 1 MiB

        private ReadOnlySpan<byte> CommittedDirectoryBytes => new[] { (byte)'/', (byte)'0', (byte)'/' };
        private ReadOnlySpan<byte> WorkingDirectoryBytes => new[] { (byte)'/', (byte)'1', (byte)'/' };
        private ReadOnlySpan<byte> SynchronizingDirectoryBytes => new[] { (byte)'/', (byte)'_', (byte)'/' };

        private U8Span CommittedDirectoryPath => new U8Span(CommittedDirectoryBytes);
        private U8Span WorkingDirectoryPath => new U8Span(WorkingDirectoryBytes);
        private U8Span SynchronizingDirectoryPath => new U8Span(SynchronizingDirectoryBytes);

        private FileSystemClient _fsClient;
        private IFileSystem _baseFs;
        private SdkMutexType _mutex;
        // Todo: Unique file system for disposal
        private int _openWritableFileCount;
        private bool _isPersistentSaveData;
        private bool _canCommitProvisionally;
        private bool _useTransactions;

        private class DirectorySaveDataFile : IFile
        {
            private IFile _baseFile;
            private DirectorySaveDataFileSystem _parentFs;
            private OpenMode _mode;

            public DirectorySaveDataFile(DirectorySaveDataFileSystem parentFs, IFile baseFile, OpenMode mode)
            {
                _parentFs = parentFs;
                _baseFile = baseFile;
                _mode = mode;
            }

            protected override Result DoRead(out long bytesRead, long offset, Span<byte> destination,
                in ReadOption option)
            {
                return _baseFile.Read(out bytesRead, offset, destination, in option);
            }

            protected override Result DoWrite(long offset, ReadOnlySpan<byte> source, in WriteOption option)
            {
                return _baseFile.Write(offset, source, in option);
            }

            protected override Result DoFlush()
            {
                return _baseFile.Flush();
            }

            protected override Result DoGetSize(out long size)
            {
                return _baseFile.GetSize(out size);
            }

            protected override Result DoSetSize(long size)
            {
                return _baseFile.SetSize(size);
            }

            protected override Result DoOperateRange(Span<byte> outBuffer, OperationId operationId, long offset, long size, ReadOnlySpan<byte> inBuffer)
            {
                return _baseFile.OperateRange(outBuffer, operationId, offset, size, inBuffer);
            }

            protected override void Dispose(bool disposing)
            {
                if (disposing)
                {
                    _baseFile?.Dispose();

                    if (_mode.HasFlag(OpenMode.Write))
                    {
                        _parentFs.DecrementWriteOpenFileCount();
                        _mode = default;
                    }
                }

                base.Dispose(disposing);
            }
        }

        public static Result CreateNew(out DirectorySaveDataFileSystem created, IFileSystem baseFileSystem,
            bool isPersistentSaveData, bool canCommitProvisionally, bool useTransactions,
            FileSystemClient fsClient = null)
        {
            var obj = new DirectorySaveDataFileSystem(baseFileSystem, fsClient);
            Result rc = obj.Initialize(isPersistentSaveData, canCommitProvisionally, useTransactions);

            if (rc.IsSuccess())
            {
                created = obj;
                return Result.Success;
            }

            obj.Dispose();
            UnsafeHelpers.SkipParamInit(out created);
            return rc;
        }

        /// <summary>
        /// Create an uninitialized <see cref="DirectorySaveDataFileSystem"/>.
        /// </summary>
        /// <param name="baseFileSystem">The base <see cref="IFileSystem"/> to use.</param>
        public DirectorySaveDataFileSystem(IFileSystem baseFileSystem)
        {
            _baseFs = baseFileSystem;
            _mutex.Initialize();
        }

        /// <summary>
        /// Create an uninitialized <see cref="DirectorySaveDataFileSystem"/>.
        /// If a <see cref="FileSystemClient"/> is provided a global mutex will be used when synchronizing directories.
        /// Running outside of a Horizon context doesn't require this mutex,
        /// and null can be passed to <paramref name="fsClient"/>.
        /// </summary>
        /// <param name="baseFileSystem">The base <see cref="IFileSystem"/> to use.</param>
        /// <param name="fsClient">The <see cref="FileSystemClient"/> to use. May be null.</param>
        public DirectorySaveDataFileSystem(IFileSystem baseFileSystem, FileSystemClient fsClient)
        {
            _baseFs = baseFileSystem;
            _mutex.Initialize();
            _fsClient = fsClient;
        }

        protected override void Dispose(bool disposing)
        {
            _baseFs?.Dispose();
            base.Dispose(disposing);
        }

        private Result Initialize(bool isPersistentSaveData, bool canCommitProvisionally, bool useTransactions)
        {
            _isPersistentSaveData = isPersistentSaveData;
            _canCommitProvisionally = canCommitProvisionally;
            _useTransactions = useTransactions;

            // Ensure the working directory exists
            Result rc = _baseFs.GetEntryType(out _, WorkingDirectoryPath);

            if (rc.IsFailure())
            {
                if (!ResultFs.PathNotFound.Includes(rc)) return rc;

                rc = _baseFs.CreateDirectory(WorkingDirectoryPath);
                if (rc.IsFailure()) return rc;

                if (_isPersistentSaveData)
                {
                    rc = _baseFs.CreateDirectory(CommittedDirectoryPath);
                    if (rc.IsFailure()) return rc;
                }

                // Nintendo returns on all failures, but we'll keep going if committed already exists
                // to avoid confusing people manually creating savedata in emulators
                if (rc.IsFailure() && !ResultFs.PathAlreadyExists.Includes(rc)) return rc;
            }

            // Only the working directory is needed for temporary savedata
            if (!_isPersistentSaveData)
                return Result.Success;

            rc = _baseFs.GetEntryType(out _, CommittedDirectoryPath);

            if (rc.IsSuccess())
            {
                if (!_useTransactions)
                    return Result.Success;

                return SynchronizeDirectory(WorkingDirectoryPath, CommittedDirectoryPath);
            }

            if (!ResultFs.PathNotFound.Includes(rc)) return rc;

            // If a previous commit failed, the committed dir may be missing.
            // Finish that commit by copying the working dir to the committed dir

            rc = SynchronizeDirectory(SynchronizingDirectoryPath, WorkingDirectoryPath);
            if (rc.IsFailure()) return rc;

            return _baseFs.RenameDirectory(SynchronizingDirectoryPath, CommittedDirectoryPath);
        }

        private Result ResolveFullPath(Span<byte> outPath, U8Span relativePath)
        {
            if (StringUtils.GetLength(relativePath, PathTools.MaxPathLength + 1) > PathTools.MaxPathLength)
                return ResultFs.TooLongPath.Log();

            U8Span workingPath = !_useTransactions && _isPersistentSaveData
                ? CommittedDirectoryPath
                : WorkingDirectoryPath;

            StringUtils.Copy(outPath, workingPath);
            outPath[outPath.Length - 1] = StringTraits.NullTerminator;

            return PathNormalizer.Normalize(outPath.Slice(2), out _, relativePath, false, false);
        }

        protected override Result DoCreateFile(U8Span path, long size, CreateFileOptions options)
        {
            Unsafe.SkipInit(out FsPath fullPath);

            Result rc = ResolveFullPath(fullPath.Str, path);
            if (rc.IsFailure()) return rc;

            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            return _baseFs.CreateFile(fullPath, size, options);
        }

        protected override Result DoDeleteFile(U8Span path)
        {
            Unsafe.SkipInit(out FsPath fullPath);

            Result rc = ResolveFullPath(fullPath.Str, path);
            if (rc.IsFailure()) return rc;

            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            return _baseFs.DeleteFile(fullPath);
        }

        protected override Result DoCreateDirectory(U8Span path)
        {
            Unsafe.SkipInit(out FsPath fullPath);

            Result rc = ResolveFullPath(fullPath.Str, path);
            if (rc.IsFailure()) return rc;

            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            return _baseFs.CreateDirectory(fullPath);
        }

        protected override Result DoDeleteDirectory(U8Span path)
        {
            Unsafe.SkipInit(out FsPath fullPath);

            Result rc = ResolveFullPath(fullPath.Str, path);
            if (rc.IsFailure()) return rc;

            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            return _baseFs.DeleteDirectory(fullPath);
        }

        protected override Result DoDeleteDirectoryRecursively(U8Span path)
        {
            Unsafe.SkipInit(out FsPath fullPath);

            Result rc = ResolveFullPath(fullPath.Str, path);
            if (rc.IsFailure()) return rc;

            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            return _baseFs.DeleteDirectoryRecursively(fullPath);
        }

        protected override Result DoCleanDirectoryRecursively(U8Span path)
        {
            Unsafe.SkipInit(out FsPath fullPath);

            Result rc = ResolveFullPath(fullPath.Str, path);
            if (rc.IsFailure()) return rc;

            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            return _baseFs.CleanDirectoryRecursively(fullPath);
        }

        protected override Result DoRenameFile(U8Span oldPath, U8Span newPath)
        {
            Unsafe.SkipInit(out FsPath fullCurrentPath);
            Unsafe.SkipInit(out FsPath fullNewPath);

            Result rc = ResolveFullPath(fullCurrentPath.Str, oldPath);
            if (rc.IsFailure()) return rc;

            rc = ResolveFullPath(fullNewPath.Str, newPath);
            if (rc.IsFailure()) return rc;

            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            return _baseFs.RenameFile(fullCurrentPath, fullNewPath);
        }

        protected override Result DoRenameDirectory(U8Span oldPath, U8Span newPath)
        {
            Unsafe.SkipInit(out FsPath fullCurrentPath);
            Unsafe.SkipInit(out FsPath fullNewPath);

            Result rc = ResolveFullPath(fullCurrentPath.Str, oldPath);
            if (rc.IsFailure()) return rc;

            rc = ResolveFullPath(fullNewPath.Str, newPath);
            if (rc.IsFailure()) return rc;

            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            return _baseFs.RenameDirectory(fullCurrentPath, fullNewPath);
        }

        protected override Result DoGetEntryType(out DirectoryEntryType entryType, U8Span path)
        {
            Unsafe.SkipInit(out FsPath fullPath);

            Result rc = ResolveFullPath(fullPath.Str, path);
            if (rc.IsFailure())
            {
                UnsafeHelpers.SkipParamInit(out entryType);
                return rc;
            }

            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            return _baseFs.GetEntryType(out entryType, fullPath);
        }

        protected override Result DoOpenFile(out IFile file, U8Span path, OpenMode mode)
        {
            UnsafeHelpers.SkipParamInit(out file);

            Unsafe.SkipInit(out FsPath fullPath);

            Result rc = ResolveFullPath(fullPath.Str, path);
            if (rc.IsFailure()) return rc;

            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            rc = _baseFs.OpenFile(out IFile baseFile, fullPath, mode);
            if (rc.IsFailure()) return rc;

            file = new DirectorySaveDataFile(this, baseFile, mode);

            if (mode.HasFlag(OpenMode.Write))
            {
                _openWritableFileCount++;
            }

            return Result.Success;
        }

        protected override Result DoOpenDirectory(out IDirectory directory, U8Span path, OpenDirectoryMode mode)
        {
            UnsafeHelpers.SkipParamInit(out directory);

            Unsafe.SkipInit(out FsPath fullPath);

            Result rc = ResolveFullPath(fullPath.Str, path);
            if (rc.IsFailure()) return rc;

            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            return _baseFs.OpenDirectory(out directory, fullPath, mode);
        }

        /// <summary>
        /// Creates the destination directory if needed and copies the source directory to it.
        /// </summary>
        /// <param name="destPath">The path of the destination directory.</param>
        /// <param name="sourcePath">The path of the source directory.</param>
        /// <returns>The <see cref="Result"/> of the operation.</returns>
        private Result SynchronizeDirectory(U8Span destPath, U8Span sourcePath)
        {
            // Delete destination dir and recreate it.
            Result rc = _baseFs.DeleteDirectoryRecursively(destPath);

            // Nintendo returns error unconditionally because SynchronizeDirectory is always called in situations
            // where a PathNotFound error would mean the save directory was in an invalid state.
            // We'll ignore PathNotFound errors to be more user-friendly to users who might accidentally
            // put the save directory in an invalid state.
            if (rc.IsFailure() && !ResultFs.PathNotFound.Includes(rc)) return rc;

            rc = _baseFs.CreateDirectory(destPath);
            if (rc.IsFailure()) return rc;

            // Lock only if initialized with a client
            if(_fsClient is not null)
            {
                using ScopedLock<SdkMutexType> lk =
                    ScopedLock.Lock(ref _fsClient.Globals.DirectorySaveDataFileSystem.SynchronizeDirectoryMutex);

                using (var buffer = new RentedArray<byte>(IdealWorkBufferSize))
                {
                    return Utility.CopyDirectoryRecursively(_baseFs, destPath, sourcePath, buffer.Span);
                }
            }
            else
            {
                using (var buffer = new RentedArray<byte>(IdealWorkBufferSize))
                {
                    return Utility.CopyDirectoryRecursively(_baseFs, destPath, sourcePath, buffer.Span);
                }
            }
        }

        protected override Result DoCommit()
        {
            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            if (!_useTransactions || !_isPersistentSaveData)
                return Result.Success;

            if (_openWritableFileCount > 0)
            {
                // All files must be closed before commiting save data.
                return ResultFs.WriteModeFileNotClosed.Log();
            }

            Result RenameCommittedDir() => _baseFs.RenameDirectory(CommittedDirectoryPath, SynchronizingDirectoryPath);
            Result SynchronizeWorkingDir() => SynchronizeDirectory(SynchronizingDirectoryPath, WorkingDirectoryPath);
            Result RenameSynchronizingDir() => _baseFs.RenameDirectory(SynchronizingDirectoryPath, CommittedDirectoryPath);

            // Get rid of the previous commit by renaming the folder
            Result rc = Utility.RetryFinitelyForTargetLocked(RenameCommittedDir);
            if (rc.IsFailure()) return rc;

            // If something goes wrong beyond this point, the commit will be
            // completed the next time the savedata is opened

            rc = Utility.RetryFinitelyForTargetLocked(SynchronizeWorkingDir);
            if (rc.IsFailure()) return rc;

            rc = Utility.RetryFinitelyForTargetLocked(RenameSynchronizingDir);
            if (rc.IsFailure()) return rc;

            return Result.Success;
        }

        protected override Result DoCommitProvisionally(long counter)
        {
            if (!_canCommitProvisionally)
                return ResultFs.UnsupportedCommitProvisionallyForDirectorySaveDataFileSystem.Log();

            return Result.Success;
        }

        protected override Result DoRollback()
        {
            // No old data is kept for temporary save data, so there's nothing to rollback to
            if (!_isPersistentSaveData)
                return Result.Success;

            return Initialize(_isPersistentSaveData, _canCommitProvisionally, _useTransactions);
        }

        protected override Result DoGetFreeSpaceSize(out long freeSpace, U8Span path)
        {
            UnsafeHelpers.SkipParamInit(out freeSpace);

            Unsafe.SkipInit(out FsPath fullPath);

            Result rc = ResolveFullPath(fullPath.Str, path);
            if (rc.IsFailure()) return rc;

            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            return _baseFs.GetFreeSpaceSize(out freeSpace, fullPath);
        }

        protected override Result DoGetTotalSpaceSize(out long totalSpace, U8Span path)
        {
            UnsafeHelpers.SkipParamInit(out totalSpace);

            Unsafe.SkipInit(out FsPath fullPath);

            Result rc = ResolveFullPath(fullPath.Str, path);
            if (rc.IsFailure()) return rc;

            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            return _baseFs.GetTotalSpaceSize(out totalSpace, fullPath);
        }

        internal void DecrementWriteOpenFileCount()
        {
            using ScopedLock<SdkMutexType> lk = ScopedLock.Lock(ref _mutex);

            _openWritableFileCount--;
        }
    }
}
